import random
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt

def single_manhattan(filenames, chr_col, pos_col, p_col, variant_col, label_vec, color_vec, x_size, y_size, y_scale_break, y_scale_padding, p_threshold, y_label_vec, title_str, locus_labels, output_path):
    f,ax = plt.subplots(1,1,sharex=True, facecolor='w', gridspec_kw={'hspace': 0})
    f.set_size_inches(x_size, y_size)
    all_max = 0
    trait_num = 0
    for filename in filenames:
        df = pd.read_csv(filename, sep = "\t")
        df['log_p'] = -np.log10(df[p_col])
        if variant_col != None:
            chr_col = 'CHR'
            pos_col = 'POS'
            df[chr_col] = df[variant_col].apply(lambda x: 23 if x.split(":")[0] == "X" else int(x.split(":")[0]))
            df[pos_col] = df[variant_col].apply(lambda x: int(x.split(":")[1]))
        
        if trait_num == 0:
            x_labels = []
            x_labels_pos = []
            chrom_offset = {}
            chrom_start = {}
            chrom_offset_num = 0
            for chrom in range(1,23):
                chrom_offset[chrom] = chrom_offset_num
                chrom_start[chrom] = min(df[df[chr_col] == chrom][pos_col])
                x_labels.append(str(chrom))
                x_labels_pos.append(chrom_offset_num + (float(max(df[df[chr_col] == chrom][pos_col])) 
                                                        - float(min(df[df[chr_col] == chrom][pos_col])))/2)
                chrom_offset_num += max(df[df[chr_col] == chrom][pos_col]) - min(df[df[chr_col] == chrom][pos_col])
        
        if len(filenames) > 1:
            trait_max = single_plot_multi(ax, df, chr_col, pos_col, chrom_offset, chrom_start, color_vec, label_vec, trait_num)
            if all_max < trait_max:
                all_max = trait_max
            if trait_num == 0:
                master_df = df
            else:
                master_df = pd.concat([master_df,df])
        else:
            all_max = single_plot_single(ax, df, chr_col, pos_col, chrom_offset, chrom_start, color_vec, trait_num)
            master_df = df
        trait_num += 1

    ax.set_xlim([0, chrom_offset_num])
    ax.set_ylim([y_scale_break, all_max + y_scale_padding])
    ax.set_ylabel(y_label_vec[0])
    ax.set_xlabel('Chromosome')
    plt.xticks(fontsize = 8)
    plt.yticks(fontsize = 8)

    #Set chromosome labels
    ax.set_xticks(x_labels_pos)
    ax.set_xticklabels(x_labels)
    if len(filenames) > 1: #Chromosome labels when color is used for multiple traits
        for chrom in range(1,23):
            ax.vlines(chrom_offset[chrom], ymin = 0, ymax = all_max + y_scale_padding, linestyles = 'dotted', linewidth = 1)
        ax.legend(loc = 'upper left', bbox_to_anchor= (1.01, 1.01))

    #Horizontal cutoff lines
    if p_threshold != 0:
        ax.hlines(y=-np.log10(p_threshold), xmin = 0, xmax = chrom_offset_num, linewidth = 1)
        
    #Add SNP labels
    if locus_labels != None:
        locus_labels_list = []
        label_file = open(locus_labels, 'r')
        for line in label_file:
            locus_labels_list.append(line.rsplit('\t'))
        add_locus_labels(ax, master_df, chr_col, pos_col, locus_labels_list, all_max + y_scale_padding, p_threshold)

    f.suptitle(title_str)
    
    if output_path == None:
        plt.show()
    else:
        plt.savefig(output_path, format = 'pdf', dpi = 400)

def single_plot_multi(ax, df, chr_col, pos_col, chrom_offset, chrom_start, color_vec, label_vec, trait_num):
    df['offset_num'] = df[chr_col].map(chrom_offset) - df[chr_col].map(chrom_start)
    df['IND'] = df[pos_col] + df['offset_num']
    ax.scatter(x = df['IND'], y = df['log_p'], color = color_vec[trait_num], s = 3, label = label_vec[trait_num], rasterized = True)
    return max(df['log_p'].tolist())

def single_plot_single(ax, df, chr_col, pos_col, chrom_offset, chrom_start, color_vec, trait_num):
    df['offset_num'] = df[chr_col].map(chrom_offset) - df[chr_col].map(chrom_start)
    df['IND'] = df[pos_col] + df['offset_num']
    for chrom in range(1,23):
        tempdf = df[df[chr_col] == chrom]
        ax.scatter(x = tempdf['IND'], y = tempdf['log_p'], color = color_vec[chrom%2], s = 3, rasterized = True)
    return max(df['log_p'].tolist())

def add_locus_labels(ax, master_df, chr_col, pos_col, locus_labels_list, all_max, p_threshold):
    horiz_align = ['left', 'right']
    locus_ind = 0
    vert_offset = -np.log10(p_threshold) + np.dot((all_max + np.log10(p_threshold))/len(locus_labels_list), range(0,len(locus_labels_list)))
    random.shuffle(vert_offset)
    for locus in locus_labels_list:
        df_temp = master_df[(master_df[chr_col] == int(locus[0])) & (master_df[pos_col] >= int(locus[1])) & (master_df[pos_col] <= int(locus[2]))]
        df_ind = df_temp['log_p'].argmax()
        point_coord = [int(df_temp.iloc[[df_ind]]['IND']),float(df_temp.iloc[[df_ind]]['log_p'])]
        label_coord = [int(df_temp.iloc[[df_ind]]['IND']),vert_offset[locus_ind] + (all_max + np.log10(p_threshold))/(2*len(locus_labels_list))]
        if int(locus[0]) == 1 : alignment = 'left'
        elif int(locus[0]) >= 20: alignment = 'right'
        else: alignment = horiz_align[locus_ind % 2]
        ax.annotate(locus[3].rstrip(), xy=(point_coord), xytext=(label_coord), horizontalalignment = alignment, 
                    arrowprops=dict(arrowstyle="->"), color = locus[4].rstrip())
        locus_ind += 1


import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="easyManhattan",
    version="1.0.1",
    author="James Han",
    author_email="jamesjisu@gmail.com",
    description="An easy-to-use package to generate Manhattan plots",
    long_description="An easy-to-use package to generate a diverse range of Manhattan plots, including double Manhattan plots to visualize paired traits (such as male-female stratified GWAS results).",
    long_description_content_type="text/markdown",
    url="https://github.com/jamesjisu/easyManhattan",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=["numpy","pandas","matplotlib"],
    python_requires='>=3.6',
)



import argparse
from easyManhattan import double_files
from easyManhattan import single_files

def main():
    parser = argparse.ArgumentParser(description='Make a Manhattan plot')

    parser.add_argument('--filenames', 
                        nargs = '+', 
                        required = True,
                        help = 'List of summary statistic file paths', 
                        dest = 'filenames1')
    parser.add_argument('--double-filenames', 
                        nargs = '*', 
                        help = 'List of second summary statistic file paths', 
                        dest = 'filenames2')
    parser.add_argument('--chr-col',
                        help = 'Name of the chr-value column',
                        dest = 'chr_col')
    parser.add_argument('--pos-col',
                        help = 'Name of the pos-value column',
                        dest = 'pos_col')
    parser.add_argument('--p-col',
                        required = True,
                        help = 'Name of the p-value column',
                        dest = 'p_col')
    parser.add_argument('--variant-col',
                        help = 'Name of the variant column (formatted as CHR:POS:A1:A2)',
                        dest = 'variant_col')
    parser.add_argument('--names', 
                        nargs = '+', 
                        required = True,
                        help = 'List of summary statistic trait names', 
                        dest = 'label_vec')
    parser.add_argument('--colors',
                        nargs = '+',
                        required = True,
                        help = 'List of marker colors for multiple traits, or list of 2 colors for alternating chromosome marker colors',
                        dest = 'color_vec')
    parser.add_argument('--width', 
                        default = 8,
                        type = float,
                        dest = 'x_size',
                        help = 'Desired width of the plot (in inches)')
    parser.add_argument('--height', 
                        default = 6,
                        type = float,
                        dest = 'y_size',
                        help = 'Desired height of the plot (in inches)')
    parser.add_argument('--y-break', 
                        type = float,
                        default = 0,
                        dest = 'y_scale_break',
                        help = 'Y-axis break, to increase readability of the plot')
    parser.add_argument('--y-padding', 
                        default = 2,
                        type = float,
                        dest = 'y_scale_padding',
                        help = 'Y-scale maximum padding, value added to y-axis maximum on plot scale')
    parser.add_argument('--p-threshold',
                        default = 0,
                        type = float,
                        dest = 'p_threshold',
                        help = 'Specify p-value threshold to plot a horizontal line (5e-8 usually used)')
    parser.add_argument('--y-labels', 
                        nargs = '+',
                        dest = 'y_label_vec',
                        help = 'Labels for y-axis (1 for single Manhattan plot, 2 for double Manhattan plot)')
    parser.add_argument('--plot-title', 
                        default = 'Manhattan Plot',
                        dest = 'title_str',
                        help = 'Name of the plot')
    parser.add_argument('--locus-labels',
                        dest = 'locus_labels',
                        help = 'Path to a tab-delimited file with locus labels')
    parser.add_argument('--output-path',
                        dest = 'output_path',
                        help = 'Desired output path of plot')

    args = parser.parse_args()

    #Check for valid arguments
    if (args.variant_col == None) and ((args.chr_col == None) or (args.pos_col == None)):
        parser.error('Either the name of a chromosome and position column or the name of a variant column is required')

    if len(args.filenames1) > 1:
        if len(args.filenames1) != len(args.color_vec):
            raise ValueError("Number of colors specified does not match the number of summary statistics provided")
        elif len(args.filenames1) != len(args.label_vec):
            raise ValueError("Number of summary statistics and number of trait names does not match")
    else:
        if len(args.color_vec) != 2:
            raise ValueError("Please specify two colors for alternating chromosomes on the plot")

    if (args.p_threshold < 0) or (args.p_threshold > 1):
        raise ValueError("Invalid value for p-value threshold")
    elif (args.x_size <= 0) or (args.y_size <= 0):
        raise ValueError("Dimensions must be positive values")

    if args.filenames2 == None: #Single Manhattan plot case
        if args.locus_labels != None:
            loci_labels_file = open(args.locus_labels)
            for line in loci_labels_file:
                if len(line.rsplit("\t")) != 5:
                    raise ValueError("Incorrect number of columns in locus labels file (also check for any whitespace issues)")

    else: #Double Manhattan plot case
        if len(args.filenames1) != len(args.filenames2):
            raise ValueError("Number of traits in top Manhattan plot and bottom Manhattan plot do not match")    
        if args.locus_labels != None:
            loci_labels_file = open(args.locus_labels)
            for line in loci_labels_file:
                if len(line.rsplit("\t")) != 6:
                    raise ValueError("Incorrect number of columns in locus labels file (also check for any whitespace issues)")

        

    if args.filenames2 == None:
        single_files.single_manhattan(args.filenames1, args.chr_col, args.pos_col, args.p_col, args.variant_col, args.label_vec, args.color_vec, args.x_size, args.y_size, args.y_scale_break, args.y_scale_padding, args.p_threshold, args.y_label_vec, args.title_str, args.locus_labels, args.output_path)
    else:
        double_files.double_manhattan(args.filenames1, args.filenames2, args.chr_col, args.pos_col, args.p_col, args.variant_col, args.label_vec, args.color_vec, args.x_size, args.y_size, args.y_scale_break, args.y_scale_padding, args.p_threshold, args.y_label_vec, args.title_str, args.locus_labels, args.output_path)
    
if __name__ == "__main__":
    main()